<?php

namespace Drupal\content_synchronizer\Plugin\content_synchronizer\entity_processor;

use Drupal\content_synchronizer\Base\JsonWriterTrait;
use Drupal\content_synchronizer\Processors\Entity\EntityProcessorBase;
use Drupal\content_synchronizer\Processors\Entity\EntityProcessorInterface;
use Drupal\content_synchronizer\Processors\ExportProcessor;
use Drupal\content_synchronizer\Processors\ImportProcessor;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\file\Entity\File;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'accordion' formatter.
 *
 * @EntityProcessor(
 *   id = "content_synchronizer_file_processor",
 *   entityType = "file"
 * )
 */
class FileProcessor extends EntityProcessorBase implements EntityProcessorInterface, ContainerFactoryPluginInterface {
  use JsonWriterTrait;

  const DIR_ASSETS = "assets";

  /**
   * Export assets directory path.
   *
   * @var string
   */
  protected $exportAssetsDirPath;

  /**
   * Import assets directory path.
   *
   * @var string
   */
  protected $importAssetsDirPath;

  /**
   * Constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, FileSystemInterface $file_system) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->fileSystem = $file_system;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('file_system'),
    );
  }

  /**
   * Return the data to export.
   */
  public function getDataToExport(EntityInterface $entityToExport) {
    $this->addFileToAssets($entityToExport);
    return parent::getDataToExport($entityToExport);
  }

  /**
   * Add file to assets dir.
   */
  protected function addFileToAssets(File $file) {
    $assetsDir = $this->getExportAssetsDir();

    $destination = $this->createDirTreeForFileDest(str_replace('://', '/', $file->getFileUri()), $assetsDir);

    // Copy file in destination directory.
    $this->fileSystem->copy($file->getFileUri(), $destination);
  }

  /**
   * Return the export assets path.
   */
  protected function getExportAssetsDir() {
    if (!$this->exportAssetsDirPath) {
      $writer = ExportProcessor::getCurrentExportProcessor()->getWriter();
      $this->exportAssetsDirPath = $writer->getDirPath() . '/' . self::DIR_ASSETS;
      $this->createDirectory($this->exportAssetsDirPath);
    }
    return $this->exportAssetsDirPath;
  }

  /**
   * Return the import assets path.
   */
  protected function getImportAssetsDir() {
    if (!$this->importAssetsDirPath) {
      $this->importAssetsDirPath = ImportProcessor::getCurrentImportProcessor()->getImport()->getArchiveFilesPath() . '/' . self::DIR_ASSETS;

    }
    return $this->importAssetsDirPath;
  }

  /**
   * Return the entity to import.
   *
   * @param array $data
   *   The data to import.
   * @param \Drupal\Core\Entity\EntityInterface|null $entityToImport
   *   The entity to import.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The entity to import.
   */
  public function getEntityToImport(array $data, EntityInterface $entityToImport = NULL) {
    /** @var \Drupal\file\Entity\File|null $file */
    $file = parent::getEntityToImport($data, $entityToImport);
    if ($file) {
      $assetsFile = $this->getImportAssetsDir() . '/' . str_replace('://', '/', $file->getFileUri());
      if (file_exists($assetsFile)) {

        if (strpos($file->getFileUri(), '://')) {
          [$root, $destination] = explode('://', $file->getFileUri());
          $root .= '://';
        }
        else {
          [$root, $destination] = [$file->getFileUri(), '/'];
        }

        $this->createDirTreeForFileDest($destination, $root);

        $result = copy($assetsFile, $file->getFileUri());
        if ($result) {
          return $file;
        }
      }
    }

    return NULL;
  }

}
